use std::sync::{Once, Arc};
use hyper_rustls::HttpsConnector;
use rss::Channel;
use scraper::{Html, Selector};
use html5ever::{QualName, ns, namespace_url, local_name};
use rayon::prelude::*;
use axum::{Server, Router, routing::get, http::Uri, body::Body, response::IntoResponse, extract::State};
use hyper::{Client, client::HttpConnector};

static mut IMG_SELECTOR: (Option<Selector>, Once) = (None, Once::new());

fn img_selector() -> &'static Selector {
    unsafe {
        IMG_SELECTOR.1.call_once(move || {
            IMG_SELECTOR.0 = Some(Selector::parse(r#"img[alt="thumbnail"]"#).unwrap());
        });
        IMG_SELECTOR.0.as_ref().unwrap_unchecked()
    }
}

fn update_media(item: &rss::Item) -> rss::Item {
    try_update_media(&item).unwrap_or(item.to_owned())
}

fn try_update_media(item: &rss::Item) -> Option<rss::Item> {
    // correct media URL we want to use
    let url = item.extensions().get("media")?.get("content")?.get(0)?.attrs().get("url")?.as_str();

    let html = Html::parse_fragment(item.description.as_ref()?.as_str());

    let img_ref = html.select(img_selector()).next().unwrap();
    let img_html_before = img_ref.html();
    let mut img = img_ref.value().to_owned();
    img.attrs.get_mut(&QualName::new(None, ns!(), local_name!("src"))).unwrap().clone_from(&url.into());

    let updated_html = html.root_element().inner_html().replace(&img_html_before, format!("{img:?}").as_str());

    let mut item = item.clone();
    item.description = Some(updated_html);
    Some(item)
}

#[axum::debug_handler]
async fn rss_proxy_handler(uri: Uri, State(client): State<Arc<Client<HttpsConnector<HttpConnector>>>>) -> axum::response::Result<impl IntoResponse> {
    let result = || async {
        let query = uri.query().unwrap_or_default();
        let new_uri = Uri::builder()
            .scheme("https")
            .authority("backend.deviantart.com")
            .path_and_query(format!("/rss.xml?{query}"))
            .build()?;
        
        let response_buf = hyper::body::to_bytes(client.get(new_uri).await?).await?;
        let mut rss = Channel::read_from(response_buf.as_ref())?;
        rss.items = rss.items().into_par_iter().map(update_media).collect();

        Ok::<_, Box<dyn std::error::Error>>(rss.to_string())
    };

    Ok(result().await.map_err(|err| format!("{err:#?}"))?)
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let client = Arc::new(Client::builder()
        .build::<_, Body>(hyper_rustls::HttpsConnectorBuilder::new()
            .with_native_roots()
            .https_only()
            .enable_http2()
            .build()
        ));

    let app = Router::new()
        .route("/", get(rss_proxy_handler))
        .with_state(client);

    let socket_addrs = ("0.0.0.0:80".parse()?, "0.0.0.0:8080".parse()?);
    Server::try_bind(&socket_addrs.0)
        .or_else(|_| {
            eprintln!("Could not connect to port 80, now using port 8080");
            Server::try_bind(&socket_addrs.1)
        })?
        .serve(app.clone().into_make_service())
        .await?;

    Ok(())
}
